######################################
## Systemic Coder
#
#  File: Codings.tcl
#  Author: Mick O'Donnell micko@wagsoft.com
#
#  Copyright: the author grants you a limited, nonexclusive, royalty-free 
#  right to reproduce and distribute this file; provided, however, that 
#  use of such software shall be subject to the terms and conditions of 
#  the Systemic Coder License Terms, a copy of which may be viewed at:
#  http://www.wagsoft.com/Coder/copyright.txt
#  This notice must not be removed from the software, and in the event
#  that the software is divided, it should be attached to every part.

# Codings DB Interface
# Functions dealing with Codings

global NODE CURRENT CURRINDX TEXT_NODES UNUSED_IDS LAST_ID


# NODE -- Array holding the NODE.
  #  -- uses an artificially generated key to index it.
  # NODE(1,text) - string
  # NODE(1,comment) - string
  # NODE(1,features) - list
  # NODE(1,completed-p) - 0 or 1
  # NODE(1,ignore) - 0 or 1

# TEXT_NODES -- a list of the NODE indexs in order of text.
# CURRENT -- the index into NODE of the current
# CURR_INDEX -- the index into TEXT_NODES of the current


# clears interface and all vars
proc reset-all {} {
    global GRAMMAR_FILE CODINGS_FILE TEXT_FILE 
    clear-codings
    clear-text
    clear-systems
    set GRAMMAR_FILE {}
    set TEXT_FILE {}
    set CODINGS_FILE {}
}


proc clear-codings {} {
  global CURR_INDEX CURRENT NODE TEXT_NODES LAST_ID UNUSED_IDS REVIEW_FILTER\
	  MODIF_FLAG  SCHEME_MODIFIED REVIEW_COLOR_SYSTEM SEGMENT_COLOR_SYSTEM DOCUMENT_FEATURES

  if [info exists NODE] {
    unset NODE 
  }
  set TEXT_NODES {}
  set CURR_INDEX -1
  set CURRENT {}
  set MODIF_FLAG 0
  set SCHEME_MODIFIED 0
  mark-unedited

  set LAST_ID 0
  set UNUSED_IDS {}
  set REVIEW_FILTER {}
  set REVIEW_COLOR_SYSTEM {}
  set SEGMENT_COLOR_SYSTEM {}
  set DOCUMENT_FEATURES {}
}


proc clear-node {cid} {
  global NODE
  set NODE($cid,text) {}
  set NODE($cid,textstripped) {}
  set NODE($cid,comment) {}
  set NODE($cid,features) {}
  set NODE($cid,completed-p) 0
  set NODE($cid,ignore) 0
  set NODE($cid,tmp_choices) {}
}


proc new-node {} {
  global UNUSED_IDS LAST_ID NODE
  if { $UNUSED_IDS != {} } {
	set id [pop UNUSED_IDS]
  } else {
      set id [incr LAST_ID]
  }
  clear-node $id
  return $id
}


proc new-text-node {} {
    global  NODE ROOT
    set cid [new-node]
    set NODE($cid,features) [list $ROOT]
    return $cid
}

proc nth-coding {pos} {
  global TEXT_NODES
  return [lindex $TEXT_NODES $pos]
}


proc node-position {id} {
  global TEXT_NODES
  return [lsearch -exact $TEXT_NODES $id]
}


proc forget-node {nid} {
    global UNUSED_IDS NODE TEXT_NODES CURRENT
    debug "forget-node: $nid"

    # 1. Make this nid available again
    push $nid UNUSED_IDS

    # ensure CURRENT points to a real coding
    if { $CURRENT == $nid } {
	set pos [node-position $nid] 
	if { $pos == 0 } {
	    set CURRENT [nth-coding 0]
	} else {
	    set CURRENT [nth-coding [expr $pos - 1]]
	}   
    }	    
    
    delete TEXT_NODES $nid
}

proc mark-edited {} {
    global MODIF_FLAG
    if { $MODIF_FLAG != 1} {
	set MODIF_FLAG 1
	MenuItemEnable .menuBar "Codings" "Save Codings"
    }
}

proc mark-unedited {} {
    global MODIF_FLAG SCHEME_MODIFIED
    if {$MODIF_FLAG != 0} {
	set MODIF_FLAG 0
	if { $SCHEME_MODIFIED == 0 } {
	    MenuItemDisable .menuBar "Codings" "Save Codings"
	}
    }
}


proc codings-modified-p {} {
    # I treat the coding as modified if the scheme is modified
    # regardless of whether it is external or embedded. This may 
    # cause false saves of codings if using an external scheme.
    global MODIF_FLAG SCHEME_MODIFIED
    if [info exists SCHEME_MODIFIED] {
	if { $SCHEME_MODIFIED == 1 } { return 1 }
    }
    if [info exists MODIF_FLAG] {return $MODIF_FLAG}
    return 0
}



proc ignore-coding {nid} {
    global NODE
    set NODE($nid,ignore) 1
    mark-edited
}

proc unignore-coding {nid} {
    global NODE
    set NODE($nid,ignore) 0
    mark-edited
}

proc all-unignored-codings {} {
    global NODE TEXT_NODES
    set result {}
    foreach nid $TEXT_NODES {
	if { $NODE($nid,ignore) == 0 } {
	    lappend result $nid
	}
    }
    return $result
}

proc calculate-node-completeness {nid {resolve_gates 0}} {
    # All gates should be resolved when this is called
    global NODE
    if { $resolve_gates != 0} {
	set NODE($nid,features) [select-from-gates $NODE($nid,features)]
    }
    if { [unresolved-systems $NODE($nid,features)] == {} } {
	set NODE($nid,completed-p) 1
    } else { 
	set NODE($nid,completed-p) 0
    }
}

proc calculate-all-node-completeness {{resolve_gates 0}} {
    # All gates should be resolved when this is called
    global TEXT_NODES
    foreach node $TEXT_NODES {
	calculate-node-completeness $node $resolve_gates
    }
}

