######################################
## Systemic Coder
#
#  File: Filter.tcl
#  Author: Mick O'Donnell micko@wagsoft.com
#
#  Copyright: the author grants you a limited, nonexclusive, royalty-free 
#  right to reproduce and distribute this file; provided, however, that 
#  use of such software shall be subject to the terms and conditions of 
#  the Systemic Coder License Terms, a copy of which may be viewed at:
#  http://www.wagsoft.com/Coder/copyright.txt
#  This notice must not be removed from the software, and in the event
#  that the software is divided, it should be attached to every part.


# Code for filtering codings based on feature spec
# DISPLAYED_FILTER: initially set to the displayed filter string
#                   modified to reflect wildcard replacement.

proc apply-filter {filter nodes} {
    global DISPLAYED_FILTER

    # we have a list inside a quotes
    if {[llength $filter] == 1 } {
	if [intersect-p {and or not} [car $filter]] {
	    set filter [car $filter]
	}
    }

    # Store the displayed filter
    set DISPLAYED_FILTER $filter

    set filter [delete-newlines [string trim $filter]]

    if {$filter == {} } {
	return [all-unignored-codings]
    }

    # Normalise the filter
    regsub -all {\(} $filter "\{" filter
    regsub -all {\)} $filter "\}" filter

    # we have a list inside a quotes
    # set filter [lindex $filter 0]
    
    # Normalise to reverse polish
    set filter [interpret-bool-expr $filter]

    # match
    return [filter-match $filter $nodes]
}


proc filter-match {filter nodes} {

    if { [llength $filter] > 1  } {
	set matches [complex-match $filter $nodes]
	return [lsort -integer $matches]
    } else {
	simple-match $filter $nodes
    }
}

proc simple-match {filter nodes} {
    global FEATURE NODE DISPLAYED_FILTER
    set feature [lindex $filter 0]
    if  { [string first "*" $filter] > -1 } {
	set newfeature [wildcard-match $feature feature]
	if { $newfeature == "cancelled"} {return}
	set DISPLAYED_FILTER [string-replace $DISPLAYED_FILTER $feature $newfeature]
	set feature $newfeature
    }

    # we now have a fully specified feature -check it is defined
    if ![feature-exists-p $feature] {
	tell-user "Can't apply filter, since feature: $feature is undefined."
	return "cancelled"
    }
    
    # Go!
    set matches {}
    foreach nid $nodes {
	if { $NODE($nid,ignore) == 0 && [member $feature $NODE($nid,features)] } {
	    lappend matches $nid
	}
    }
    return $matches
}
	    

proc complex-match {filter nodes} {
    global TEXT_NODES
    switch -- [lindex $filter 0] {
	and {
	    set matches $nodes
	    foreach itm [lrange $filter 1 end] {
		if { $matches == {} || $matches == "cancelled"} { return }
		set matches [filter-match $itm $matches]
	    }
	    return $matches
	}
	or {
	    set matches {}
	    foreach itm [lrange $filter 1 end] {
		set new [filter-match $itm $nodes]
		if { $new == "cancelled"} {return "cancelled"}
		set matches [delete-duplicates [concat $matches $new]]
	    }
	    return $matches
	}
	not {
	    set matches [filter-match [lindex $filter 1] $nodes]
	    return [set-difference $nodes $matches]
	}
	    
	default {
	    tell-user "Complex Filter failed because first item should be one of: and, or, not"
	    return "cancelled"
	}
    }
}

################3
# TRANSLATE BOOLEAN to REVERSE POLISH


proc interpret-bool-expr {be} {

    switch -- [llength $be] {
	0 {}
	1 { 
	    # It is either a feature or an embedded BE
	    set feat [car $be]
	    if {[llength $feat] == 1} {
		return $feat
	    } else {
		return [interpret-bool-expr $feat]
	    }
	}
	
	2 {
	    # Has to be a not
	    if { [car $be] == "not" } {
		return [list not [interpret-bool-expr [second $be]]]
	    } else {
		tell-user "Error in filter: 2 item element: $be"
		return 0
	    }
	}
	
	default {
	    set result {}
	    
	    # Process the first item
	    set first [pop be]
	    if { $first == "not"} {
		set first [list not [interpret-bool-expr [pop be]]]
	    } else {
		set first [interpret-bool-expr $first]
	    }
	    lappend result $first
	    
	    # process the rest
	    set mixedbrackets 0
	    while { $be != {} } {
		set conj [pop be]
		if ![member $conj {and or}] {
		    tell-user "Error in filter: expected 'and' or 'or': $be"
		    return 0
		}
		lappend result $conj
		
		if [info exists lastconj] {
		    if { $conj != $lastconj} {
			set mixedBrackets 1
		    }
		}
		
		set next [pop be]
		
		if { $next == {} } {
		    tell-user "Error in filter: expected elelemnt after '$conj'"
		    return 0
		}
		
		if { $next == "not"} {
		    set next [list not [interpret-bool-expr [pop be]]]
		} else {
		    set next  [interpret-bool-expr $next]
		}
		
		lappend result $next
		
		set lastconj $conj
	    }
	    
	    # We are now finished
	    return [bool-bracket-elements  $result]
	}
    }
}


proc bool-bracket-elements {blist} {
   # Converts a (a and b or c) list to (or (and a b) c) list
    set elements [list [pop blist]]
    set conj [pop blist]
    lappend elements [pop blist]

    while { $blist != {} } {
	
	if { [pop blist] == $conj } {
	    # We have a continuation, add the next element
	    lappend elements [pop blist]
	} else {
	    # We have a change of conj
	    switch -- $conj {
		and {
		    # We had an and now hit an or
		    # So treat all elements so far as one, 
		    # See what the rest of the list is,
		    # if it is an or, then add this conj as the first element
		    # if it is and and return (this-conj or result)
		    set thisitem [concat and $elements]
		    if { [llength $blist] == 1} {
			return [list or $thisitem [car $blist]]
		    }
		    set rest [bool-bracket-elements $blist]
		    switch -- [car $rest] {
			and {
			    return [list or $thisitem $rest]
			}
			or {
			    return [list or $thisitem [cdr $rest]]
			}
			
			default {
			    # The list had a single entity
			}
		    }
		}
		or {
		    # We had an 'or' and hit an 'and'.
		    # Take the last element and all until we hit an 'or', or end,
		    # and add the and of these  as next element.
		    # Then continue if elements left.
		    set conjlist [list [pop-from-end elements]]
		    lappend conjlist [pop blist]
		    while { $blist != {} && [pop blist] == "and"} {
			
			# Add the next conj
			lappend conjlist [pop blist]
		    }

		    # End of the and run, add the conj to the disjunct list    
		    lappend elements [concat and $conjlist]
		}
	    }
	}
    }
    return [concat $conj $elements]
}


