######################################
## Systemic Coder
#
#  File: Load.tcl
#  Author: Mick O'Donnell micko@wagsoft.com
#
#  Copyright: the author grants you a limited, nonexclusive, royalty-free 
#  right to reproduce and distribute this file; provided, however, that 
#  use of such software shall be subject to the terms and conditions of 
#  the Coder License Terms, a copy of which may be viewed at:
#  http://www.wagsoft.com/Coder/copyright.html
#  This notice must not be removed from the software, and in the event
#  that the software is divided, it should be attached to every part.


# General Functions to do with Files

proc default-file {{ext "cd3"}} {
    global CODINGS_FILE TEXT_FILE DIR ANALYSES
    if { $CODINGS_FILE != {} } {
	set default $CODINGS_FILE
    } elseif  {$TEXT_FILE != {} } {
	set default $TEXT_FILE
    } elseif [file isdirectory $ANALYSES] {
	set default [file join $ANALYSES "untitled"]
    } else {
	set default [file join $DIR "untitled"]
    }
    return "[file rootname $default].$ext"
}

proc file-load-types {} {
    global FILE_LOAD_FORMS
    set types {}
    foreach type $FILE_LOAD_FORMS {
	lappend types [list [car $type] [second $type]]
    }
    return $types
}


proc add-load-option {title ext proc} {
    global FILE_LOAD_FORMS 
    # If this type found, override
    if [info exists FILE_LOAD_FORMS] {
	indexed-delete FILE_LOAD_FORMS $title 
    }

    # Add the entry
    lappend FILE_LOAD_FORMS [list $title $ext $proc]
}


#######################################
## LOAD

proc load-codings {{file {}}} {
    global CODINGS_FILE CODINGSFORMAT TEXT_FILE TEXT_NODE  DIR CURRENT CURRENT_INTERFACE\
	    DS_FILE FILE_LOAD_FORMS

      
    if { [prompt-save-current-files] == "cancelled"} {
	return 0
    }

    if { $CURRENT_INTERFACE != "segment"} {
	uninstall-interface
    }

    if { $file == {} } {
	    # Compile out the load formats
	set types {}
	foreach type $FILE_LOAD_FORMS {
	    lappend types [list [car $type] [second $type]]
	}
	set default [default-file "cd3"]
	set initfile [file tail $default]
	set initdir [file dirname $default]
	set file [tk_getOpenFile  -filetypes $types\
		-initialfile  $initfile -initialdir $initdir]

	if {$file == {}} {return}
    }

    if ![file exists $file] {
	tell-user "File not found: $file"
	return 0
    }

    set CODINGS_FILE $file

    #1. wipe the grammar
    clear-codings
    clear-text
    set DS_FILE {}
    
    set extension [string tolower [file extension $CODINGS_FILE]]
    foreach type $FILE_LOAD_FORMS {
	if  [member $extension [second $type]] {

	    # Loads the codings, displaying a watch cursor
	    start-watch-cursor
	    if [catch "call-file-loader [third $type] \"$CODINGS_FILE\"" result] {
		stop-watch-cursor
		BugReport "Error in loading $CODINGS_FILE. Check file and if ok send to micko@wagsoft.com with following details:"
		return 0
	    }
	    stop-watch-cursor
	    if { $result == 0} {return 0}
	    break
	}
    }
    
    
    # Show the text
    if {$extension != ".res1"} {
	display-text
    }


    # Layout/Draw the net
    graph-scheme
    set CURRENT 1
    if {$extension == ".res1"} {
	install-interface scheme
    } else {
	install-interface segment
    }
    mark-unedited
    install-stats-mode descriptive
    wm title . [file tail $CODINGS_FILE]
}

proc call-file-loader {proc file} {
    return [eval "$proc \"$file\""]
}


# Codings: cd2

proc load-xml-forms  {file} {
    global ENCODING 
    # Loads in the xml file. Reloads if encoding specified in file
    
    load-xml $file 
    set codingsform [parse-xml {system encoding}]

    #####################################
    # Check for international encoding
    set encoding [string tolower [get-xml-element $codingsform {codings header encoding name}]]
    if { $encoding != {} } {
	# Call the function again but this time read in encoded mode.
	set ENCODING $encoding
	load-xml $file $encoding
	set codingsform [parse-xml {system encoding}]
    } else {
	if [info exists ENCODING] {unset ENCODING}
    }
    return $codingsform
}


    
proc load-codings2  {file {encoding {}}} {
    global ROOT
    # Expects a heirarchically structured xml form.
    # Only Segment tags can include free text.

    set codingsform [load-xml-forms $file]

    ###############################
    # parse the scheme
    clear-systems
    parse-scheme-xml [get-xml-element $codingsform {codings header scheme}]

    ###############################
    # parse any document features (features of a document rather than a segment - used in doc comparisons)
    add-doc-features [get-xml-elements $codingsform {codings header docfeature}]


    ################################
    # PARSE the body
    process-body [get-xml-element $codingsform {codings body}]
    return 1
}


proc add-doc-features {forms} {
    global DOCUMENT_FEATURES
    set DOCUMENT_FEATURES {}
    foreach form $forms {
	set val [second [car [second $form]]]
	lappend DOCUMENT_FEATURES $val
    }
}


proc verify-scheme-location {scheme_file} {
    global  SCHEMES CODINGS_FILE
    if [file exists $scheme_file] {
	return $scheme_file
    }
    
    ## 1. Check if the file is in the Relations DIR.
    set fname [file tail $scheme_file]
    set altpath [file join $SCHEMES $fname]
    if [file exists $altpath] {
	switch -- [ask-user "The recorded scheme file was not found at:\n   $scheme_file.\n  Should I use the following instead?\n  $altpath"] {
	    yes { return $altpath }
	}
    }

    # Try in the same directory as the CODINGS File
    set altpath [file join [file dirname $CODINGS_FILE] $fname]
    if [file exists $altpath] {
	switch -- [ask-user "The recorded Relations file was not found at:\n   $scheme_file.\n  Should I use the following instead?\n  $altpath"] {
	    yes { return $altpath }
	}
    }
    
    # No Alternatives - offer load/search
    switch -- [tk_dialog .relfind "Scheme file location"\
	    "Scheme: $scheme_file has been moved." \
	    {} 0  Locate "Try Without" "Cancel Load"] {
	0 { 
	    set newfile [tk_getOpenFile -initialfile [file tail $scheme_file]\
		    -initialdir [file dirname $CODINGS_FILE]]
	    if { $newfile != {} } {
		return $newfile
	    } else {
		return "cancelled"
	    }
	}
	1 {  return {} }
	
	2 { return "cancelled"} 
    }
}



proc process-body {body} {
    global NODE ROOT TEXT_NODES 

    #3. Load in segment defins
    foreach seg [get-xml-elements $body {body segment}] {

	# Features
	set features [get-xml-element $seg {segment features}]

	# Text
	set text [xml-decode [lindex $seg 3]]
	
	# Comment
	set comment [xml-decode [get-xml-element $seg {segment comment}]]

	# Ignore
	if { [get-xml-element $seg {segment ignore}] == 1} {
	    set ignore  1
	} else {
	     set ignore 0
	}

	add-text-node $text $features $comment $ignore
    }
}


proc strip-coder-text {text} {
    regsub -all "\n" [string trim $text] " " text
    regsub -all "\[ \t\]*\[ \t\]" $text " " text
    return $text
}

proc parse-xml-system {system} {
    set name     [get-xml-element $system {system name}]
    set ec       [get-xml-element $system {system ec}]
    set features [get-xml-element $system {system features}]
    set state    [get-xml-element $system {system state}]
    set gloss    [get-xml-element $system {system gloss}]
    defsystem $name $ec $features $state
    if { $gloss != {} } {
	defgloss [string toupper $name] $gloss
    }
}

proc parse-xml-feature {feature} {
    set name     [get-xml-element $feature {feature name}]
    set realisations       [get-xml-element $feature {feature realisations}]
    # set gloss    [get-xml-element $feature {feature gloss}]
    set state    [get-xml-element $feature {feature state}]
    regsub -all ";" $realisations "\n" realisations
    def-feature $name $realisations $state
}

proc parse-scheme-xml {schemedef} {
    global GRAMMAR_FILE

    ###############################
    # parse the scheme
    set scheme_file [get-xml-element $schemedef {scheme file}]

    if { $scheme_file == {} } {
	# Assume scheme is embedded
	set GRAMMAR_FILE {}
	defroot [get-xml-element $schemedef {scheme root}]
	foreach system [get-xml-elements $schemedef {scheme system}] {
	    parse-xml-system $system
	}
	foreach feature [get-xml-elements $schemedef {scheme feature}] {
	    parse-xml-feature $feature
	}
    } else {
	# Scheme is in separate file
	# Handle relocated files
	set scheme_file [verify-scheme-location $scheme_file]
	if {$scheme_file == {} } {
	    tell-user "No Scheme file: loading without."
	    set GRAMMAR_FILE {}
	    create-scheme
	} elseif {$scheme_file == "cancelled"} {
	    return 0
	} else {
	    load-scheme $scheme_file
	}
    }

    # Ensure root is defined
    if {[info exists ROOT] == 0 || $ROOT == {} } {
	# Root not yet defined
	find-root
    }
}



proc load-scheme-from-codings-file {{file {}} {graphp 0}} {
    global GRAMMAR_FILE XMLSEG LASTSEG CURRSEG ROOT DOCUMENT_FEATURES

    # We do this here in case the scheme breaks on loads
    # and leaves us with a corrupt scheme which the user might save
    set GRAMMAR_FILE {}

    clear-systems

    ##################################
    #  Parse in the xml form
    set text [read-file $file]
    set start [string first "<scheme" $text]
    set end [string first "</scheme>" $text]
    set text [string range $text $start [expr $end + 8]]
    load-xml-from-text $text

    parse-scheme-xml [parse-xml {system encoding}]

    foreach feature $DOCUMENT_FEATURES {
	add-doc-feature $feature
    }

    # Update the codings for the new scheme
    calculate-all-node-completeness resolve_gates

    # Redraw
    graph-scheme
}


proc copy-scheme-from {} {
    set default [default-file "cd3"]
    set initfile [file tail $default]
    set initdir [file dirname $default]
    set file [tk_getOpenFile -title "Select Codings file with Scheme to copy"\
	    -initialfile $initfile -initialdir $initdir\
	    -filetypes {{"CODINGS 2.0/3.0" {.cd2 .cd3}}}]
    if { $file == {}} {return}
    load-scheme-from-codings-file $file
    apply-imported-scheme
    mark-scheme-modified
    mark-edited
    graph-scheme
}


proc apply-imported-scheme {} {
    global TEXT_NODES NODE ORDERED_FEATURES

    if ![info exists ORDERED_FEATURES] {
	build-ordered-features-list
    }
    foreach nid $TEXT_NODES {
	set features $NODE($nid,features)
	# Abandon features which dont exist now
	set features [intersection  $ORDERED_FEATURES $features]
	set newpath [definite [path $features]]
	if { $newpath == {} } { 
	    # The path is now inconsistent
	    # order most delicate first

	    set features [lreverse $features]
	    set newpath [definite [path [pop features]]]
	    set features [set-difference $features $newpath]
	    while { $features != {}} {
		set feature [pop features]
		set path [path [concat and $feature $newpath]]
		if { $path != {}} {
		    set newpath [definite $path]
		    set features [set-difference $features $newpath]
		}
	    }
	}
	set NODE($nid,features) [select-from-gates $newpath]
	set NODE($nid,features) [select-from-gates $NODE($nid,features)]
	calculate-node-completeness $nid
    }
}

proc load-res1-file {file} {
    console show
    set scheme_file [tk_getOpenFile -title "Scheme File" -filetypes {{{Scheme file} .scheme}}]
    if {$scheme_file == {}} {return}
    load-scheme $scheme_file

    set str [open $file r]
    set Nodes {}
    while { [gets $str line] > -1 } {
	if { $line == {}} {continue}
	set firstcolon [string first ":" $line]
	set key [string range $line 0 [expr $firstcolon -1]]
	set line [string range $line [expr $firstcolon + 2] end]
	
	switch -- $key {
	    "TEXT" {
		add-res1-node $Nodes 
		set Nodes {}
	    }
	    
	    "TOKENS" {
		set Tokens {}
		foreach token [split $line "#"] {
		    regsub -all "%HATCH" $token "#" token
		    lappend Tokens $token
		}
	    }
	    
	    "ANALYSES" {
		while { [gets $str edge] > -1 } {
		    if { $edge == ""} {break}
		    # puts "Adding edge: $edge"
		    regsub -all " // " $edge "\n" edge
		    regsub " //" $edge "\n" edge
		    set line [split $edge "\n"]
		    set features [lindex $line 1]
		    set pos [lindex $line 3]
		    set token [lindex $Tokens $pos]
		    # puts "TOK: /$token/ Line: $edge"
		    lappend Nodes [list $pos $features $token]
		}
	    }

	    default {
		puts "****UNKNOWN $key"
	    }
	}
    }
    close $str
    
    # Store the last item
     add-res1-node $Nodes 
}

proc add-res1-node {nodes} {
    # puts "adding nodes: $nodes"
    set nodes [lsort -index 0 -integer $nodes]
    foreach node $nodes {
	add-text-node "[third $node] " [second $node] "" 0 0
    }
}

proc add-text-node {text features comment ignore {calcCompleteness 1}} {
    global NODE ROOT TEXT_NODES
    set id [new-node]
    
    # Features
    set features1 {}
    foreach feature $features {
	    if [feature-exists-p $feature] {
		lappend features1 $feature
	    }
	}
    if {$features1 == {}} {set features1 [list $ROOT]}
    set NODE($id,features) $features1

    if {$calcCompleteness == 1} {
	set NODE($id,features) [select-from-gates $NODE($id,features)]
	calculate-node-completeness $id
    }

    set NODE($id,text) $text
    set NODE($id,textstripped) [strip-coder-text $text]

    # Comment
    set NODE($id,comment) $comment
    set NODE($id,ignore)  $ignore
    lappend TEXT_NODES $id
}