###################################3
# Documenter.tcl

# Code  to provide auto documentation of systems and features

proc document-system {system} {
    global SYSTEM FEATURE

    start-doc-result
    add-doc "<p><hr><h2><a name=\"#$system\">System: $system</h2><hr><p>"

    # provide the context of the system
    set features $SYSTEM($system,features)
    set ec $SYSTEM($system,ec)
    switch -- [llength $features] {
	1 {
	    set type gate
	}
	
	2 { 
	    set type binary
	}

	default {
	    set type multiple
	}
    }

    switch -- $type {
	gate {
	    add-doc "This is a gate, a system consisting of a single feature. "

	    if [simple-ec-p $ec] {
		add-doc "Normally gates have a complex entry condition, but this one does not, "
		add-doc "it thus states a simple equivalence between  '$ec' and '$features'. "
	    } else {

		# Select the quantifier
		set word [list-quantifier [car $ec] [cdr $ec]]
		set list [comma-list [cdr $ec] [car $ec]]
		add-doc "Its single feature: '$features' is selected "
		add-doc "when $word of the features $list are selected. " 
	    }
	}

	default {

	    if [simple-ec-p $ec] {
		add-doc "This system represents a sub-classification of the feature: '$ec'. "
		add-doc "'$ec' units thus need to be  "
		if { [llength $features] == 2 } {
		    add-doc "either [car $features] or [second $features]. "
		} else {
		    add-doc "one of: [comma-list $features or]. "
		}
	    } else {
		set word [list-quantifier [car $ec] [cdr $ec]]
		set eclist [comma-list [cdr $ec] [car $ec]]
		set flist  [comma-list $features [car $ec]]
		add-doc "This system is entered if $word of $eclist is selected. "
		add-doc "When this is so, the unit must also select one of $flist. "
	    }
	}
    }
    
    return [return-doc-result]
}


proc document-feature {feature} {
    global ROOT FEATURE PRESEL_LOOKUP

    if [concept-p $feature] {
	return [document-concept $feature]
    }

    start-doc-result
    set extenders  $FEATURE($feature,extenders)

    # Provide the contex of the feature
    if { $feature == $ROOT } {
	add-doc "<p>The root of the network is the feature: '$ROOT'."
    } else {
	add-doc "<p><h3>Feature: $feature</h3><p>"
    }

    if ![info exists FEATURE($feature,realisations)] {
	add-doc "The feature: '$feature' has no associated realisation statements. "
    } else {
	add-doc "It has the following realisation: <ul>"
	foreach real [split $FEATURE($feature,realisations) "\n"] {
	    add-doc "<li> $real"
	}
	add-doc "</ul>"
    }

    if [info exists PRESEL_LOOKUP($feature)] {
	set refs $PRESEL_LOOKUP($feature)
	add-doc "\n\nThe feature is referenced in [llength $refs] preselections: \n\n"
	foreach ref $refs {
	    set presel [third $ref]
	    if [complex-ec-p $presel] {
		set presel "{$presel}"
	    }
	    
	    add-doc "  \[[car $ref]\] -> [second $ref]: $presel\n"
	}
    } else {
	add-doc "\n\nThe feature is not referenced in any preselections.\n"
    }
    
    add-doc "\nThis feature has "
    switch -- [llength $extenders] {
	0 { add-doc "no dependent systems. " }
	1 { add-doc "one dependent system: '$extenders'." }
	2 { add-doc "two dependent systems: [comma-list $extenders and]." }
	default {
	    add-doc "several dependent systems: [comma-list $extenders and]."
	}
    }

    return [return-doc-result]
}

    
proc start-doc-result {} {
    global DOC_STRING
    set DOC_STRING ""
}

proc add-doc {string} {
    global DOC_STRING
    set DOC_STRING "$DOC_STRING$string"
}

proc return-doc-result {} {
    global DOC_STRING
    return $DOC_STRING 
}


proc document-all {} {
    global DONE ROOT QUEUE SYSTEM
    set DONE {}
    set QUEUE {}

    puts "\n\n==========================================================\n\n"
    puts "[document-from-feature $ROOT]"

    while { $QUEUE != {} } {
	set system [pop QUEUE]
	puts "[document-system $system]"
	lappend DONE $system
	set first 1
	foreach feature $SYSTEM($system,features) {
	   document-from-feature $feature   
	}
    }
    puts "\n\n---------end----------\n"
}

proc document-from-feature {feature} {
    global FEATURE QUEUE DONE
   
    foreach system $FEATURE($feature,extenders) {
	if { [member $system $DONE] == 0 && [member $system $QUEUE] == 0 } {
	    push $system  QUEUE
	}
    }
    return [document-feature $feature]
}


proc tmp {} {
    "ENTRY:
If cross-classification: This system represents a simultaneous choice
with one other system, <other system>.
cross-classifi"
}


proc comma-list {items conj} {
    switch -- [llength $items] {
	0 {return ""}
	1 {return [car $items] }
	default {
	    set result "'[pop items]'"
	    while { [llength $items] > 1 } {
		set result "$result, '[pop items]'"
	    }
	    return "$result $conj '[car $items]'"
	}
    }
}


proc list-quantifier {conj items} {
    # Select the quantifier
    switch -- $conj {
	and { 
	    if { [llength $items] == 2 } {
		return "both"
	    } 
	    return "all"
	}
	or {
	    if { [llength $items] == 2 } {
		return "either"
	    } 
	    return "any"
	}
    }
}

proc show-system-documentation {system} {
    global SYSTEM
    set text [document-system $system]
    foreach feature $SYSTEM($system,features) {
	set text "$text\n[document-feature $feature]"
    }
    set text [html-to-text $text]
    
    if [winfo exists .sysdoc] {destroy .sysdoc}
    return [scrolled-text-dialog-nonmodal .sysdoc "Description of system: $system"  $text  -height 18 -font 14]
}

proc html-to-text {text} {
    regsub -all "<hr>" $text "\n=======================================\n" text
    regsub -all "<p>" $text "\n" text
    regsub -all "<h3>" $text "" text
    regsub -all "<h2>" $text "" text
    regsub -all "</h3>" $text "" text
    regsub -all "</h2>" $text "" text
    regsub -all "</h2>" $text "" text
    regsub -all "</a>" $text "'" text
    regsub -all "<li>" $text "\n  * " text
    regsub -all "<ul>" $text "\n" text
    regsub -all "</ul>" $text "\n\n" text
    regsub -all "<a href=.*?>" $text "'" text
    regsub -all "<a name=.*?>" $text "" text
    return $text
}

proc show-feature-documentation {feature} {
    set text [document-feature $feature]
      set text [html-to-text $text]
    scrolled-text-dialog .featdoc "Description of feature: $feature"  $text  -height 18
}


proc document-all-html {file} {
    global DONE ROOT QUEUE SYSTEM GRAMMAR_FILE BG_COLOR1
    set DONE {}
    set QUEUE {}

    set str [open $file w]

    puts $str "<html><head><title>Documentation for $GRAMMAR_FILE</title></head>\n"
    puts $str "<body bg=\"$BG_COLOR1\">\n"
    puts $str "\n<center><h1><hr>Documentation of Scheme: $GRAMMAR_FILE</h1>"
    puts $str "\n<h2>File date: [date]<hr></h2></center>"
    set text [document-from-feature $ROOT]

    while { $QUEUE != {} } {
	set system [pop QUEUE]
	set text "$text[document-system $system]"
	lappend DONE $system
	set first 1
	foreach feature $SYSTEM($system,features) {
	   set text "$text[document-from-feature $feature]" 
	}
    }

    
    
    puts $str "\n$text\n</body></html>"
    close $str
}

proc url-ref {feature} {
    return "<a href=\"#$feature\">$feature</a>"
}

proc url-ref-conjunction {features} {
    set string [url-ref [car $features]]
    foreach feature [cdr $features] {
	set string "$string, [url-ref $feature]"
    }
}

proc document-concept {concept} {
    global ROOT CONCEPT PRESEL_LOOKUP FEATURE

    start-doc-result

    # Provide the contex of the concept
    add-doc "<hr><a name=\"#$concept\"><h3>Concept: $concept</h3><hr><p>"

    if { $concept == "domain-entity" } {
	add-doc "This concept is the root of the Linkbase ontology."
    } else {
	set parent [fetch-concept-parents $concept]
	if  [simple-ec-p $parent] {
	    add-doc "'$concept' is a subtype of concept: [url-ref $parent]."
	} else {
	    add-doc "'$concept' is a subtype of several concepts: [url-ref-conjunction $parent]. "
	}
    }

    if [info exists FEATURE($concept,realisations)] {
	add-doc "\nIt has the following criteria:<ul>"
	foreach real [split $FEATURE($concept,realisations) "\n"] {
	    add-doc "<li> $real"
	}
	add-doc "</ul>"
    } else {
	add-doc "\nThe concept: '$concept' has no associated criteria. "
    }

    if [info exists PRESEL_LOOKUP($concept)] {
	set refs $PRESEL_LOOKUP($concept)
	add-doc "<p>The concept is referenced in [llength $refs] preselections: <ul>"
	foreach ref $refs {
	    set presel [third $ref]
	    if [complex-ec-p $presel] {
		set presel "{$presel}"
	    }
	    
	    add-doc "<li>\[[car $ref]\] -> [second $ref]: $presel\n"
	}
	add-doc "</ul>"
    } else {
	add-doc "The concept is not referenced in any preselections. "
    }
    
    set extenders  $FEATURE($concept,extenders)
    add-doc "\nThis concept has "
    switch -- [llength $extenders] {
	0 { add-doc "no dependent systems. " }
	1 { add-doc "one dependent system: '$extenders'." }
	2 { add-doc "two dependent systems: [comma-list $extenders and]." }
	default {
	    add-doc "several dependent systems: [comma-list $extenders and]."
	}
    }

    return [return-doc-result]
}

# regsub -all {#.*&} text "'" text
# regsub -all "#(.*)&" $text {'\1'} text1
 
# set fred "this is #finite& and not."
