# New Grapher


## COPING WITH Variable bracket size
#
# Curly brackets need to be different widths depending on the vertical height
# Layout needs to be able to recover this width.
# So, for each system, store the width of the pre-bracket, which is 0
# for systems which are sole extenders.
# We can also add the width of a system's complex ec to this measure.

# Problem: we only draw a curly bracket if there is at least two
# systems adjacent to the feature. If we have complex ecs, then one
# extending system may in fact occur elsewhere in the diagram.
# solution:
#   a) for each system with a simple ec, it is bracketed
#   b) for each system with a complex ec, delay its drawing until the rest 
#      of the graph is drawn. However, the curly bracket has 2 textlines
#      assigned for the system (a line will start there and go to the system
#   c) The systems with a complex ec are drawn

# Easier: always have a bracket if there is more than one visible extender
# Add the distance at that point, and calculate the space needed.

register-grapher "Normal Graph" draw-graph1

proc draw-graph1 {feature} {
    global YMAX XMAX VISITED
    init-graph $feature
    set YMAX 10
    set XMAX 0
    
    layout-feat $feature 10 
    
    if [info exists VISITED] {unset VISITED}
    show-feature $feature 

    fix-grapher-scrollbars
}



#######################################################
## LAYOUT

# Determines the screen coords of each feature/system

proc layout-feat {feature xpos} {
    global YMAX SPOS FPOS XINCR YINCR FEATURE XMAX \
            FEAT_FONT FHEIGHT SCREEN_XMAX RHEIGHT REALISATION_FONT
   
    set FPOS($feature,x) $xpos
    set FPOS($feature,postspace) 0
    set FPOS($feature,xend) [expr $xpos + $FPOS($feature,width)]
    set FPOS($feature,cbn) 0 ;# curly bracket needed - defaults to not.
    set child_x [expr [end-of-feature $feature] + $XINCR]

    if { $child_x > $XMAX} {
	set XMAX $child_x
	# Hack to make scroll work
	set SCREEN_XMAX [expr $child_x + $FPOS($feature,width)]
    }

    set ymax 0
    set ymin 999999
    foreach sys $FEATURE($feature,extenders) {

	# If this system hidden, ignore
	if ![info exists SPOS($sys,wgt)] {continue}
	
	if [info exists SPOS($sys,y)] {
	    # sys has a position, so increment its X if needed.
	    if { $child_x > $SPOS($sys,x) } {
		set SPOS($sys,x) [expr $child_x + $SPOS($sys,prespace)]
		set SPOS($sys,xend) [expr $SPOS($sys,x) + $SPOS($sys,width)]
	    }
	} else {
	    layout-sys $sys $child_x
	    set ymax [max $SPOS($sys,ymid) $ymax]
	    set ymin [min $SPOS($sys,ymid) $ymin]
	}
    }
    
    # Now,  YLAYOUT this feature
    if { $ymax > 0 } {
	set FPOS($feature,y) [expr $ymin + (( $ymax - $ymin) / 2)]
	
	# That places the top of the feature at the middle of the system, correct
	# to get the middle at the middle
	set FPOS($feature,y) [expr $FPOS($feature,y) - ($FPOS($feature,height) / 2)]
	set FPOS($feature,yend) [expr $FPOS($feature,y) + $FPOS($feature,height)]
	
	# Adjust for curly brackets
	if [curly-bracket-needed $feature] {
	    
	    set xshift [min 40 [expr (($ymax -$ymin) / 13) + 16] ]
	    
	    set FPOS($feature,postspace) $xshift
	    foreach system $FEATURE($feature,extenders) {
		# If this system hidden, ignore
		if [info exists SPOS($system,wgt)] {
		    replace-subnet $system
		}
	    }
	}
    } 

    # Check if the feature wasn't laid out
    if ![info exists FPOS($feature,y)] {
	set FPOS($feature,y) $YMAX
	set FPOS($feature,yend) [expr $FPOS($feature,y) +  $FPOS($feature,height)]
	set YMAX [expr $FPOS($feature,yend) + $YINCR]
    }
    
    # Now we are stable, set this value
    set FPOS($feature,ymid) [expr $FPOS($feature,y) + ($FPOS($feature,height) / 2)]

    # if this feature has realisations, allow room
    set reals [fetch-feature-realisations $feature]
    if { $reals != {}} {
	set needed [expr $FPOS($feature,yend) + $FPOS($feature,rheight) + $YINCR +3]
	set YMAX [max $YMAX $needed]
    }
}

proc curly-bracket-needed {feature} {
    global FEATURE SPOS SYSTEM FPOS
    # 1 See if curly brackets are needed
    # Only use curly bracket if there is at least one system with simple ec


    set count 0
    foreach sys $FEATURE($feature,extenders) {
	# If this system hidden, ignore
	if [info exists SPOS($sys,wgt)] {
	    if ![complex-ec-p $SYSTEM($sys,ec)] {
		incr count
	    }
	}
    }
    
    
    if { $count > 1} {
	set FPOS($feature,cbn) 1
    } else {
	set FPOS($feature,cbn) 0
    }
    return $FPOS($feature,cbn)
}
    

proc replace-subnet {sys} {
    global SPOS SYSTEM FPOS XINCR FEATURE

    # Some feature has changed position or postspace, 
    # so re-do x layout for subnet

    set xmax 0
    foreach feat $SYSTEM($sys,entry_features) {
	if [info exists FPOS($feat,xend)] {
	    set xpos [expr [end-of-feature $feat] + $FPOS($feat,postspace) + $XINCR]
	    set xmax [max $xmax $xpos]
	}
    }

    incr xmax $SPOS($sys,prespace)
    if { $xmax != $SPOS($sys,x) } {
	# xpos has changed
	set SPOS($sys,x) $xmax
	set SPOS($sys,xend) [expr $xmax + $SPOS($sys,width)]
	set featstart [expr $SPOS($sys,xend) + $XINCR]
	foreach feat $SYSTEM($sys,features) {
	    set FPOS($feat,x) $featstart
	    set FPOS($feat,xend) [expr $featstart + $FPOS($feat,width)]
	    foreach sys1 $FEATURE($feat,extenders) {
		# If this system hidden, ignore
		if [info exists SPOS($sys1,wgt)] {
		    replace-subnet $sys1
		}
	    }
	}
    }
}



proc layout-sys {system xpos} {
    global YMAX SPOS FPOS XINCR YINCR SYSTEM SYSTEM_YRANGE SYS_FONT SHEIGHT
    set prespace 0

    if [complex-ec-p $SYSTEM($system,ec)] {
	incr prespace 15
    }

    incr xpos $prespace
    set SPOS($system,prespace) $prespace
    set SPOS($system,x) $xpos 
    
    set SPOS($system,xend) [expr $xpos + $SPOS($system,width)]
    set child_x [expr $SPOS($system,xend) + $XINCR]
    set ymax 0
    set ymin 999999
    foreach feature $SYSTEM($system,features) {
	layout-feat $feature $child_x
	set ymax [max $FPOS($feature,ymid) $ymax]
	set ymin [min $FPOS($feature,ymid) $ymin]
    }
    # Now,  YLAYOUT this system
    if { $ymax > 0 } {
	set SPOS($system,y) [expr $ymin + ( $ymax - $ymin) / 2]
	
	# This has placed the top of the system where the mid should be, correct it
	set SPOS($system,y) [expr $SPOS($system,y) - ($SPOS($system,height) / 2)]
	
	set SYSTEM_YRANGE($system,min) $ymin
	set SYSTEM_YRANGE($system,max) $ymax
	
    } else { # this is a leaf
	set SPOS($system,y) $YMAX
	set YMAX [expr $SPOS($system,y) + $SPOS($system,height) + $YINCR]
  }

  # Now we are stable, calculate these handy values
  set SPOS($system,yend) [expr $SPOS($system,y) + $SPOS($system,height)]
  set SPOS($system,ymid) [expr $SPOS($system,y) + ($SPOS($system,height) / 2)]

}



#######################################################
## SHOW-FEATURE/SYSTEM

# Moves text to final position, and draws connecting lines

proc show-feature {feature} {
    global VISITED FPOS FEATURE SPOS GCANV SHEIGHT ROOT SYSTEM

    set color "black"
    if { $feature != $ROOT } {
	if { $SYSTEM($FEATURE($feature,system),state) == "disabled"} {
	    set color "gray"
	    $GCANV  itemconfigure $FPOS($feature,wgt) -fill $color
	}
    }
    $GCANV coords $FPOS($feature,wgt) $FPOS($feature,x) $FPOS($feature,y)
   
    # Move the Realisations
    if {[fetch-feature-realisations $feature] != {}} {
	$GCANV coords $FPOS($feature,rwgt) \
		[expr $FPOS($feature,x) + 5] [expr $FPOS($feature,yend) + 3]
	$GCANV itemconfigure $FPOS($feature,rwgt)  -fill $color

	# draw a rectangle around
	set bbox [$GCANV bbox $FPOS($feature,rwgt)]
	draw-rect $GCANV [expr [lindex $bbox 0] -1]  [expr [lindex $bbox 1] -1]\
		[expr [lindex $bbox 2] + 1]  [expr [lindex $bbox 3] +1]
    }
    
    # display the extending systems
    set bracketedSystems {}
    set nonbracketedSystems {}
    foreach sys $FEATURE($feature,extenders) {

	if ![info exists SPOS($sys,wgt)] {continue}
	if [info exists VISITED($sys)] {
	    draw-link $feature $sys
	} else {
	    show-system $sys
	    set VISITED($sys) 1
	    
	    # See if this system is adjacent
	    if { $SPOS($sys,x) < [expr [end-of-feature $feature] + 60] } {
		lappend bracketedSystems $sys
	    } else {
		lappend nonbracketedSystems $sys
		draw-link $feature $sys
	    }
	}
    }
    
    if { $nonbracketedSystems != {} } {
	# tell-user "nonbracketedSystems: $nonbracketedSystems"
    }
    
    # Now, draw a curly bracket if needed
    if { $FPOS($feature,cbn) == 0 } {
	foreach system $bracketedSystems {
	    draw-link $feature $system
	}
    } else {
	# multiple systems, bracket them
	set xmax 999999
	set ymin 999999
	set ymax 0
	foreach sys $bracketedSystems {
	    set xmax [min $xmax [expr $SPOS($sys,x) - $SPOS($sys,prespace)]]
	    set ymin [min $ymin $SPOS($sys,y)]
	    set ymax [max $ymax $SPOS($sys,yend)]
	}
	
	set xmin [expr [end-of-feature $feature] + 5]
	set ymin [expr $ymin - $SHEIGHT]
	set ymin [expr $ymin - $SHEIGHT]
	incr ymax $SHEIGHT
	draw-open-curly-bracket $GCANV $xmin $ymin $xmax $ymax
    }
}

proc end-of-feature {feature} {
    global FPOS
    set xpos $FPOS($feature,xend)
    if [info exists FPOS($feature,rwidth)] {
	set xpos [max $xpos [expr $FPOS($feature,x) + $FPOS($feature,rwidth) +5]]
    }
    return $xpos
}


proc show-system {system} {
    global SYSTEM SYSTEM_YRANGE SPOS FPOS GCANV
    if { $SYSTEM($system,state) == "disabled"} {
	set color "gray"
	$GCANV itemconfigure $SPOS($system,wgt) -fill $color
    } else {
	set color "black"
    }

    # if a single line, shift the system up a half-width
    $GCANV coords $SPOS($system,wgt)  $SPOS($system,x) $SPOS($system,y)
    if { $SYSTEM($system,string) == $system} {
	$GCANV move $SPOS($system,wgt) 0 [expr - ($SPOS($system,height) / 2)]
    }

    set features $SYSTEM($system,features)
    if { $features == {} } {return}
    set feature [car $features]
    set xmid [expr $FPOS($feature,x) -6]

    # Draw any complex ec
    if [complex-ec-p $SYSTEM($system,ec)] {
	# to be completed
	draw-ec $system
    }

    # Draw the system underline
    draw-line $GCANV\
                    $SPOS($system,x) $SPOS($system,ymid)\
                    $xmid $SPOS($system,ymid) "-fill $color"

    # Draw the system up bar
    # the +1 here is to fix a minor display error

    if [info exists SYSTEM_YRANGE($system,min)] {
	draw-line $GCANV\
		$xmid $SYSTEM_YRANGE($system,min)\
		$xmid $SYSTEM_YRANGE($system,max)\
		"-fill $color"
    }

    # Draw each feature
    foreach feature $features {
	show-feature $feature
	draw-line $GCANV\
                     [expr $FPOS($feature,x) -2] $FPOS($feature,ymid)\
		     $xmid $FPOS($feature,ymid)\
		     "-fill $color"
    }
}

   


proc draw-link {feature system} {
    global FPOS SPOS GCANV SYSTEM EC_HEIGHT
    if { $SYSTEM($system,state) == "disabled"} {
	set color "gray"
    } else {
	set color "black"
    }
    
    set xend [expr $SPOS($system,x) - $SPOS($system,prespace)]

    # If we end at a complex ec, work out the end pos
    if [complex-ec-p $SYSTEM($system,ec)] {
	# Work out in what pos the feature is in the ec
	set pos [screen-yorder $feature $SYSTEM($system,entry_features)]
	set count [llength $SYSTEM($system,entry_features)]
	# Divide EC_HEIGHT by count+1 to get the interval size
	# Multiply by pos+1 to get the distance from the top,
	# then subtract half EC_HEIGHT to place it relative to the middle
	set interval [expr $EC_HEIGHT / ($count + 1)]
	set offset [expr $interval * ( $pos + 1) - ($EC_HEIGHT / 2)]
	set yend [expr $SPOS($system,ymid) + $offset]
    } else {
	set yend $SPOS($system,ymid)
    }

    draw-line $GCANV\
       [expr $FPOS($feature,xend) + 2]  $FPOS($feature,ymid)\
       $xend $yend "-fill $color"
}

proc screen-yorder {feature features} {
    # Works out where in the list the feature is ordered in terms of screen y
    global FPOS
    if ![info exists FPOS($feature,y)] {
	# Feature not drawn yet
	return [llength $features]
    } 
    set ypos $FPOS($feature,y)
  
    set pos 0
    foreach feature2 [ldelete $features $feature] {
	if { [info exists FPOS($feature2,y)] == 1 \
		&& $ypos > $FPOS($feature2,y) } {incr pos}
    }
    return $pos
}

proc draw-ec {system} {
    # only one level at present
    global SHEIGHT SPOS SYSTEM GCANV REALISATION_FONT EC_HEIGHT EC_FONT
    
    switch -- [lindex $SYSTEM($system,ec) 0] {
	and { set Symb "\}" }
	or { set Symb "\]" }
    }
    set x [expr $SPOS($system,x) - $SPOS($system,prespace)]
    set y [expr $SPOS($system,ymid) - ( $EC_HEIGHT / 2 ) -5]
    draw-text $GCANV $Symb $x $y "-font [fetch-font "Times" $EC_HEIGHT]"
}


#########################
## FONT STUFF 


proc fetch-font {family size {style normal}} {
   global FONT
    switch -- $style {
	bold  {set slant roman; set weight bold}
	italic {set slant italic; set weight normal}
	default {set slant roman; set weight normal}
    }
   if ![info exists FONT($family,$size,$style)] {
       set FONT($family,$size,$style)\
	       [font create fnt-$family-$size-$style -family $family -size $size\
	       -slant $slant -weight $weight]

       # Use the font somewhere so it gets loaded and stays loaded.
       entry .tmp-$family-$size-$style -font $FONT($family,$size,$style)
   }      
   return $FONT($family,$size,$style)
}


proc string-width {str font} {

  if { [string first \n $str] == -1 } {
    return [font measure $font $str ]
  }

  set max -999
  foreach line [split $str "\n"] {
    set max [max $max [font measure $font $line]]
  }
  return $max
}


proc font-height {font} {
    return [font actual $font -size]
}



proc split-system-name {str} {
#  split the string, at the closest hyphen to the middle

   set mid_pos [expr [string length $str] / 2]
   set before_hyp_pos [string last "-" [string range $str 0 $mid_pos]]
   set after_hyp_pos [string first "-" [string range $str $mid_pos end]]

   if { $before_hyp_pos > 0 } {
      if { $after_hyp_pos != -1
           && $after_hyp_pos < [expr $mid_pos - $before_hyp_pos - 2] } {
        set mid_pos [expr  $mid_pos + $after_hyp_pos + 1]
      } else {
        set mid_pos [expr $before_hyp_pos + 1]
      }
   } else {

      if { $after_hyp_pos != -1 } {
        set mid_pos [expr $mid_pos + $after_hyp_pos + 1]
      } else {
        set mid_pos 0
      }
    }
  
    if $mid_pos {
      return [append result [string range $str 0 [expr $mid_pos -1]]\
             \n [string range $str $mid_pos  end]]
    } else {
      return $str
    }
}



