############################3333
# PATHS.tcl

# CODED BY MICK O'DONNELL
# TRANSLATED FROM CODE FROM WAG (1989-2000)

# This file provides logic for feature descriptions represented in Kasper Normal Form
# (See Kasper's article).
# A knf-path has the form:   (:and <definite> Indefinite*) 
#  where:  Definite     : a list of features (shared by all paths)
#          Indefinite: a set of FDs which are disjoint.

proc path {le} {
    global FEATURE SYSTEM ROOT

    if { [string index $le 0] == "\{" } {
	# could have used "end-1" here, but that doesnt work in tcl 8.1 or below
	set le [string trim $le]
	set butlast [expr [string length $le] -1]
	set le [string range [string trim $le] 1 $butlast]
    }

    debug "PATH: $le Type: [ec-type $le]"

    switch -- [ec-type $le] {

	fd {
	    tell-user "path: Error: called with an fd structure as arg.: $le"
	    xxx
	}
	list {
	    return [path [concat "and" $le]]
	}

	simple {
	    return [feature-path $le]
	}

	and {
	    set path [path [lindex $le 1]]
	    foreach le1 [lrange $le 2 end] {
		set path [unify-fds $path [path $le1]]
		if { $path == {} } {return {}}
	    }
	    debug "  AND path: $le  result: $path"
	    return $path
	}
	
	or {
	    set results {}
	    foreach le1 [cdr $le] {
		lappend results [path $le1]
	    }
	    disjoin-many-fds $results
	}

	not {
	    switch -- [llength $le] {
		1 {
		    tell-user "path: operator 'not' with no arguments: $le"
		    error
		}
		
		2 { return [negate-le [second $le]]}

		default  {
		    tell-user "path: operator {not} with more than one argument: $le"
		    error
		}
	    }
	}
    }
}


proc reset-all-paths {} {
    global SYSTEM
    foreach elem [array names SYSTEM "*,path"] {
	unset SYSTEM($elem)
    }
}

proc feature-path {feat} {
    global ROOT FEATURE
    if { $feat == $ROOT} {
	return [make-fd $feat {}]
    }
    set fd [system-path $FEATURE($feat,system)]
# debug "feature-path: $feat // $fd"
    set defs [concat [definite $fd] $feat]
    return [make-fd $defs [indefinites $fd]]
}

proc system-path {system} {
    global SYSTEM
    if ![info exists SYSTEM($system,path)] {
	set path [path $SYSTEM($system,ec)]
	set SYSTEM($system,path) [simplify-fd $path]	
    }
    if { $SYSTEM($system,path) == {} } { 
	tell-user "System: $system has no path"
    }
    return $SYSTEM($system,path)
}
    

proc disjoin-fds {fd1 fd2} {
    disjoin-many-fds [list $fd1 $fd2]
}

proc disjoin-many-fds {fds} {

    debug "  disjoin-many-fds $fds"

    # Take the common features from the definites of all,
    # And form what is left as an indefinite item.
    # If there an item who has no definite left,
    # and it has only one indefinite, raise the indefinite to the same level as 
    # the main indefinite item.
    #  If there is more than one, could multiply out an join??

    # Ensure all fds maximally simple (thus only one disjunction in the indef list)
    set fds1 {}
    foreach fd $fds {
	lappend fds1 [simplify-fd $fd]
    }

    set fds [delete-nulls $fds1]
    switch -- [llength $fds] {
	0 { return {}}
	1 { return [car $fds] }
    }


    # Find the common elements
    set common [definite [car $fds]]
    foreach fd [cdr $fds] {
	set common [intersection $common [definite $fd]]
    }

   debug "   common: $common"

    # if one item completely covered by the common, just return it: (:or A (:and A B)) => A
    foreach fd $fds {
	if { [indefinites $fd] == {} } {
	    if [set-equal $common [definite $fd]] {
	debug "    One Common element"
		return $fd
	    }
	}
    }

    debug "    CREATING NEW FDS extracting out the common elements"
    # Create new fds extracting out the common elements
    set Indef {}
    foreach fd $fds {
	set definite [definite $fd]
	set unique [set-difference $definite $common]

	# Indefs is a list of lists of fds. (a list of disjunctions of fds)
	# Since we expanded out each fd, there is only one disjunct list per fd
	set indefs  [indefinites $fd]

#	debug "   unique: $unique Indefs: $indefs"
	## if no definite add the only indefinite to the new units indef
	if { $unique == {} } {
	    set Indef [concat $Indef [car $indefs]]
	} else {
	    lappend Indef [make-fd $unique $indefs]
	}
    }
    
    set Indef [delete-nulls $Indef]

debug "    HERE: def: $unique INDEFS: $Indef"
    
    # delete superordinate items
    set Indef [delete-supersets $Indef]
    
debug "    After deling superordinates: def: $unique INDEFS: $Indef"

#    debug "  INDEF: $Indef"
    switch -- [llength $Indef] {
	0 { set result [make-fd $common {}] }
	
	1 {
	    ##Indef is a single item - thus definite - raise one level
	    set common [concat $common [definite [car $Indef]]]
	    set Indefs [indefinites [car $Indef]]
	    set result [make-fd $common $Indefs]
	}
	
	default {
	    # multiple indefinites - leave alone.
	    set result [make-fd $common [list $Indef]]
	}
    }

    debug "  ==disjoin-may-fds: result: $result"
    return $result
}
      

proc unify-fds {fd1 fd2} {
    # Returns a single FD being the unification of the two argument FDs.
    # note: only need to check indefs against the new definite features 
    # provided by the other path
    # so do the indef processing in two steps
    
    debug "  unify-fds fd1: $fd1 fd2: $fd2"

    ## if one fd is null, return nil
    if { $fd1 == {} || $fd2 == {} } {
	debug "    unify-fds: passed nil value: \nFD1: $fd1 \nFD2 $fd2"
	return {}
    }

    set defs1 [definite $fd1]
    set defs2 [definite $fd2]

    # only proceed if the definites unify
    if { [feature-union $defs1 $defs2] == {} } {return {}}

    ## On each iteration, take the definite features of the other FD which are not
    ## part of the current FD, and incorporate them into the FD.
    ## At the end, the definites of both FDs should be the same.

    set uniques [set-difference $defs2 $defs1]
    debug "    Unique feats in fd2: $uniques"
    
    while 1 {

	if { $uniques != {} } {
	    
	    set fd1 [unify-feats-with-fd $uniques $fd1]
	    debug "    Fd1 after taking this into account: $fd1"
	    if { $fd1 == {} } {return}
	    set defs1 [definite $fd1]
	}

	set uniques [set-difference $defs1 $defs2]
	debug "    Unique feats in fd1: $uniques"
	if { $uniques == {} } {break}
	set fd2 [unify-feats-with-fd $uniques $fd2]
	debug "    Fd2 after taking this into account: $fd2"
	if { $fd2 == {} } {return}
	set defs2 [definite $fd2]
	set uniques [set-difference $defs2 $defs1]
	debug "    Unique feats in fd2: $uniques"
	if { $uniques == {} } {break}
    }
   debug "    ALL definites incorporated!"
    set result [make-fd $defs1 [delete-nulls [concat [indefinites $fd1]\
	    [indefinites $fd2]]]]
    debug "    Returning: $result"
    return $result
}


proc definite {fd} {
#    debug "definite: fd: $fd"
    if { [car $fd] != "fd"} {
	tell-user "definite: called with argument not a knf path: $fd"
	xxx
    }

    return [lindex $fd 1]
}

proc indefinites {fd} {
    if { [car $fd] != "fd"} {
	tell-user "indefinites: called with argument not a knf path: $fd"
	xxx
    }

    return [lindex $fd 2]
}

proc make-fd {defs indefs} {
    return [list "fd" $defs $indefs]
}

proc feature-union {set1 set2} {
    
    # returns the union of the two sets, unless there
    # are features in the two sets which are mutex
    set common [intersection $set1 $set2]
    set unique1 [set-difference $set1 $set2]
    set unique2 [set-difference $set2 $set1]
    if [compat-sets $unique1 $unique2] {
	return [concat $common $unique1 $unique2]
    }
    return {}
}


proc delete-supersets {fds} {
    ## takes a disjoint list of FDs,
    ## eliminates any which is subsumed by another
    ## e.g. (or (and a b) (and a b c))  => (and a b)
    if { $fds == {} } {return {} }

#    debug "  delete-supersets: $fds"
    set head [pop fds]
    set fds2 {}
    foreach fd $fds {
	if ![knf-subsetp $head $fd] {
	    lappend fds2 $fd
	}
    }
    foreach fd $fds2 {
	if [knf-subsetp $fd $head] {
	    return [delete-supersets $fds2]
	}
    }
    return [concat [list $head] [delete-supersets $fds2]]
}

proc knf-subsetp {fd1 fd2} {
#    debug "knf-subset-p: fd1: $fd1"
#    debug "knf-subset-p: fd2: $fd2"
    if ![subset-p [definite $fd1] [definite $fd2]] {
	return 0
    }

    if { [indefinites $fd1] == {} } {
	return 1
    }

    if [knf-disjunction-subsumed-p [car [indefinites $fd1]] [car [indefinites $fd2]]] {
	return 1
    }
    
    return 0
}


proc knf-disjunction-subsumed-p {fdSet1 fdSet2} {
    ## need to find a subsumer for each of the first set in the second set
    foreach fd1 $fdSet1 {
	set subsumed 0
	foreach fd2 $fdSet2 {
	    if [knf-subsetp $fd1 $fd2] {
		set subsumed 1
	    }
	}
	if { $subsumed == 0 } {return 0}
    }
    return 1
}


#:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# UNIFICATION routines: these routines deal witht the manipulation
# FDs (Feature Descriptions), in Kasper Normal Form (KNF)
#:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::


proc unify-feats-with-fd {features fd} {
    # Returns an FD with the features propagated throughout,
    # The new features are included in the FD's definite. Note that the propagated 
    # features are included in the definite of the returned item. May have to be deleted.
    # Process: 
    # 1) Check that the new features are compat. with the definite
    # 2) Incorporate the features into each element of the Indefinites
    #   Approach: for each disjunct,resolves each alternative in respect to the definite.
    #   Three cases:
    #     0 disjuncts left:   Return nil for the whole
    #     1 disjunct left :   Unify that disjunct's def with the group definite,
    #                         collect the disjuct's indefs as fellow indefs.
    #     2 or more left:     Collect the re-worked indef.
    # 3) If any new definites have been produced, these have to be propagated also.
    #   Call the routine again with these features
    # 4) Return the definites and Indefinites.
    
    # This function works iteratively - each cyle new definite features are collected
    # and applied to the indefinites
    
    # returns nil if any indefinite fails
    
    debug "  unify-feats-with-fd: $features / fd: $fd"
    set definites [feature-union $features [definite $fd]]
    debug "    New Definites (after feature-union): $definites"
    if  { $definites == {} } { return }
    set newDefs {}
    set newIndefs {}
    foreach disjunction [indefinites $fd] {
	set newDisjunction [uf-verify-disjunction $disjunction $features]
	
	# If the disjunction failed, fail the whole
	if { $newDisjunction == 0 } {
	    return {}
	}
	
	switch -- [llength $newDisjunction] {
	    0 {
		# The disjunct was subsumed totally by $features. - do nothing
	    }
	    
	    1 {
		# Single disjunct suceeds - shift up one level
		## But first, check if the new defs are mutually compat (else fail)
		set liftedFD [car $newDisjunction]
		if ![compat-sets [definite $liftedFD] $newDefs] {
		    return {}
		}
		
		set newDefs [concat $newDefs [definite $liftedFD]]
		set newIndefs [concat $newIndefs [indefinites $liftedFD]]
	    }
	    
	    default {
		lappend newIndefs $newDisjunction
	    }
	}
    }

    debug "    --newIndefs at end: $newIndefs"
    # Now, return the results
    if { $newDefs != {} } {
	debug "    // NEWDEFS= $newDefs"
	set newStruct [unify-feats-with-fd $newDefs [make-fd {} $newIndefs]]
	if { $newStruct != {} } {
	    return [make-fd [concat $definites [definite $newStruct]]\
		    [indefinites $newStruct]]
	}
	return {}
    }
    debug "    // returning [make-fd $definites $newIndefs]"
    return [make-fd $definites $newIndefs]
}

proc require-le {le fd} {
    set fd2 [path $le]
    return [unify-fds $fd $fd2]
}

proc uf-verify-disjunction {disjunction features} {
    debug "uf-verify-disjunction: $disjunction Features: $features"
    set newDisjunction {}
    foreach fd $disjunction {
	debug "    CHECKING fd with is compat with $features FD: $fd"
	set fd [check-fd-with-features $fd $features]
	
	if {  $fd == {}} {
	    debug "    -- fd failed. Going to next"
	    continue
	}
	
	# Beware that check-fd-with-features may have returned {fd {} {}}
	# (if the features are the deminite of the fd)
	# Special case -- the disjunct is subsumed totally by $features
	# So that means we just accept the whole disjunction
	# (on the grounds that (a and (a or b)) simplifies to "a"  )
	if { $fd == [make-fd {} {}] } {
	    # This is not a fail
	    debug "    -- fd totally subsumed. Returning empty but valid result"
	    return {} 
	}
	
	# Otherwise there is an fd
	debug "    -- fd is now $fd"
	if { [definite $fd] == {} } {
	    
	    # shift the indefs up
	    debug "    -- no definites, so shifting indefs up"
	    debug "  --fd before simplification: $fd"
	    set fd [simplify-fd $fd]
	    debug "  --fd after simplification: $fd"
	    set newDisjunction [concat $newDisjunction [car [indefinites $fd]]]
	} else {
	    debug "    -- adding fd to newDisjunction list"
	    lappend newDisjunction $fd
	}
    }
    debug "    -- Indefs are now: $newDisjunction"
    
    if { $newDisjunction == {} } {
	# All disjuncts became invalid,
	# The fd as a whole is this invalid
	return 0
    }
    
    return $newDisjunction
}



proc check-fd-with-features {fd features} {

    set newFeatures [set-difference $features [definite $fd]]
    if { $newFeatures != {} } {
	set fd [unify-feats-with-fd $newFeatures $fd]
    }
    if { $fd != {} } {
	return [make-fd [set-difference [definite $fd] $features] [indefinites $fd]]
    }
}

proc simplify-fd {fd} {

    if { $fd == {} } {return {} }

    # If no indefs, then no work to do
    set indefs [indefinites $fd]
    if { $indefs == {} } { return $fd }

    set def [definite $fd]

    # Multiply out the disjoint-sets into a single disjunction of FDs
    # then simplify each component FD
    set indefs [multiply-out-indefinites $indefs]
    set newIndef {}
    foreach fd $indefs {
	set fd [simplify-fd $fd]
	if { $fd != {} } {
	    lappend newIndef $fd
	}
    }

    set newFd [disjoin-many-fds $newIndef]
    if { $newFd == {} } { return {} }

    return [make-fd [concat $def [definite $newFd]] [indefinites $newFd]]
}



proc multiply-out-indefinites {disjunctions} {
    
    set disjunct1 [pop disjunctions]
    while { $disjunctions != {} }  {
	set disjunct2 [pop disjunctions]
	set disjunct1 [fd-disjunction-expand $disjunct1 $disjunct2]
	if { $disjunct1 == {} } { return {} }
    }
    return $disjunct1
}
	
proc fd-disjunction-expand {disjunct1 disjunct2} {
    set result {}
    foreach fd1 $disjunct1 {
	foreach fd2 $disjunct2 {
	    set newFd [unify-fds $fd1 $fd2]
	    if { $newFd != {} } {
		lappend result $newFd
	    }
	}
    }
    return $result
}


proc quick-check-feature-compat {feature fd} {
    set defs [definite $fd]
    if [member $feature $defs] {
	return 1
    }

    if [intersect-p $defs [co-features $feature]] {
	return 0
    }

    set fdef [definite [feature-path $feature]]
    puts "Fdef: $fdef"
    puts "Fdef after: [set-difference $fdef $defs]"
    foreach feat [set-difference $fdef $defs] {
	if { $feat != $feature} {
	    if [intersect-p $defs [co-features $feat]] {
		return 0
	    } else {
		puts "no intersection: $defs // [co-features $feat]"
	    }
	}
    }

    # returns 1 as we have found no problems
    return 1
}

proc require-feature {feature fd} {

    set defs [definite $fd]
    if [member $feature $defs] {
	return $fd
    }

    if [intersect-p $defs [co-features $feature]] {
	return {}
    }

    set fpath [feature-path $feature]
    set fdef [definite $fpath]
    foreach feat [set-difference $fdef $defs] {
	if { $feat != $feature} {
	    if [intersect-p $defs [co-features $feat]] {
		return {}
	    } 
	}
    }

    # returns 1 as we have found no problems
    return [unify-fds $fpath $fd]
}
