############################
# Static tests

# test the grammar for consistency

proc perform-static-tests {} {
    global VALTEXT
    if [winfo exists .statictests] {destroy .statictests}
    set VALTEXT  "\n PERFORMING STATIC TESTS\n========================\n"
    # Check feature connectivity
    check-feature-connectivity
    check-realisations
    check-justified-gates
    check-realisation-combinations
    scrolled-text-dialog-nonmodal .statictests "Validation of Grammar" $VALTEXT
}

proc statictest-insert  {text} {
    global VALTEXT
    set VALTEXT "$VALTEXT\n$text"
}

proc check-feature-connectivity {} {
    global SYSTEM FEATURE ROOT
    statictest-insert "\nChecking System Connectivity:"
    statictest-insert "=============================\n"
    set found 0
    foreach elem [array names SYSTEM "*,features"] {
	set system [car [split $elem ","]]
	foreach feature $SYSTEM($system,entry_features) {
	    if { $feature == $ROOT} {continue}
	    if ![info exists FEATURE($feature,system)] {
		set found 1
		statictest-insert "   System: $system has nondefined feature: $feature in entry cond."
	    }
	}
    }

    if { $found == 0 } {
	statictest-insert "   - All features in entry conditions are defined."
    }

    statictest-insert "\nChecking System Paths:"
    statictest-insert "=============================\n"
    set found 0
    foreach elem [array names SYSTEM "*,features"] {
	set system [car [split $elem ","]]
	if { [system-path $system] == {} } {
	    statictest-insert "   System: $system has invalid path."
	}
    }
     if { $found == 0 } {
	statictest-insert "   - All system have valid paths."
    }
}


proc check-realisations {} {
    global FEATURE ST_ASSIGNS ST_INSERTS
    statictest-insert "\nChecking Realisation:"
    statictest-insert "=============================\n"
    set found 0

    if [info exists ST_ASSIGNS] {unset ST_ASSIGNS}
    if [info exists ST_INSERTS] {unset ST_INSERTS}
    foreach indx [array names FEATURE "*,realisations"] {
	set feature [car [split $indx ","]]
	set wc [feature-word-class $feature]
	if {$wc == {} } {
	    statictest-insert "   Feature: $feature has realisations but no wordclass!"
	    set found 1
	}

	set reals  [convert-reals $FEATURE($indx)]
	foreach real $reals {
	    switch -- [car $real] {
		type {
		    # This is a class assigment
		    if { [llength $real] != 3} {
			statictest-insert "   Realisation of feature: '$feature' has wrong arg count: $real"
			set found 1
		    } else {
			set funct [second $real]
			set feats [car [third $real]]
			lappend ST_ASSIGNS($wc,$funct) [list $feature $feats]

			foreach feat [flatten $feats] {
			    if ![feature-exists-p $feat] {
				statictest-insert "   Undefined feature: '$feat' in realisation of feature: '$feature'."
				statictest-insert "      Realisation:  \"$real\""
				set found 1
			    }
			}

			# Report if the presel is malformed
			if ![valid-le $feats] {
			    statictest-insert "      Malformed Preselection:  \"$funct: $feats\""
			    set found 1
			}
		    }
		}

		insert {
		    # this is a slot-assigment realisation
		    if { [llength $real] != 4} {
			statictest-insert "   Realisation of feature: '$feature' has wrong arg count: '$real'"
			set found 1
		    } else {
			set funct [second $real]
			set slot [lindex $real 2]
			set repeated [lindex $real 3]

			# Test if the slot is a valid form
			if ![member [string index $slot 0] {"+" "-"}] {
			    statictest-insert "   Slot Insert invalid:  $feature : $funct $slot\n"
			    set found 1
			    continue
			} 
			if ![string is integer [string range $slot 1 end]] {
			    statictest-insert "   Slot Insert invalid: $feature : $funct $slot\n"
			    set found 1
			    continue
			}
			    
			lappend Inserts($wc) [list $feature $funct $slot]
			lappend Inserts($wc,$slot) [list $feature $funct  $repeated]
			lappend ST_INSERTS($wc,$funct) [list $feature $slot $repeated]
			if { $slot < 0 } {
			    lappend Slots($wc,preslots) $slot
			} else {
			    lappend Slots($wc,postslots) $slot
			}
		    }
		}

		default {
		    statictest-insert "   Undefined realisation type: [car $real] in realisation of feature: $feature."
		    statictest-insert "      Realisation:  \"$real\""
		    set found 1
		}
	    }
	}
    }

    # Now run through the collections and check for compound errors
    
    #1. Check for multiple compat. inserts of a function
    foreach idx [array names ST_INSERTS] {
	set idx1 [split $idx ","]
	set wc [car $idx1]
	set funct [second $idx1]
	set inserts $ST_INSERTS($idx)
	set insert1 [pop inserts]
	while { $inserts != {}} {
	    set feat1 [car $insert1]
	    foreach insert2 $inserts {
		set feat2 [car $insert2]
		set path [simplify-path [newpath [list and $feat1 $feat2]]]
		if { $path != {} } {
		    statictest-insert "   Multiple Inserts of same function possible in one path:"
		    statictest-insert "    a) $feat1 : $funct [second $insert1]"
		    statictest-insert "    b) $feat2 : $funct [second $insert2]\n"
		    # statictest-insert "    Path: [path-to-fd $path]\n"
		    set found 1
		}
	    }
	    set insert1 [pop inserts]
	}
    }
    
    if { $found == 0 } {
	statictest-insert "   - All realisations ok."
    }
}



proc valid-le {le} {
    switch -- [car $le] {
	and -
	or {
	    if { [llength $le] < 3} {
		puts "Malformed: and $le"
		return 0
	    }
	    foreach item [cdr $le] {
		if ![valid-le $item] {
		    return 0
		}
	    }
	}
	
	not {
	    if { [llength $le] != 2} {
		puts "Malformed: not $le"
		return 0
	    }
	    if ![valid-le [second $le] {
		return 0
	    }
	}
	
	default {
	    if { [llength $le] > 1} {
		puts "Malformed: def $le"
		return 0
	    }
	    if {[string first " " $le] > -1} {
		puts "Malformed: wc $le"
		return 0
	    }
	    
	    if {[string first "\{" $le] > -1} {
		puts "Malformed: brack $le"
		return 0
	    }
	}
    }
     return 1
}

proc check-justified-gates {} {
    global SYSTEM FEATURE PRESEL_LOOKUP
    statictest-insert "\nChecking Gates are justified:"
    statictest-insert "(the single feature has realisations, subclasses, or is mentioned in a realisation)\n"
    foreach system [all-systems] {
	if { [llength $SYSTEM($system,features)] == 1} {
	    
	    set feature [car $SYSTEM($system,features)]
	    # Justified if the feature has realisations
	    if [info exists FEATURE($feature,realisations)]  {continue}
	    
	    # Justified if there are extenders
	    if {$FEATURE($feature,extenders) != {} }  {continue}
	    
	    # Justified if the feature is used in realisations
	    if [info exists PRESEL_LOOKUP($feature)] {continue}
	    
	    statictest-insert "Gate: $system has no justifiaction."
	}
    }
}


proc check-realisation-combinations {} {
    global WORD_CLASSES ST_ASSIGNS ST_INSERTS FUNCTIONS INSERTCOND PATH
    
    statictest-insert "\nChecking Combinations of Preselections are legal:"
    foreach wc $WORD_CLASSES {
	foreach idx [array names ST_ASSIGNS "$wc,*"] {
	    set funct [second [split $idx ","]]
	    
	    # 1. Get the insert condition for the function
	    if ![info exists ST_INSERTS($wc,$funct)] {
		statictest-insert "ERROR: no insert for preselected function: $funct"
		set startpath [feature-path $wc]
	    } else {
		set startpath {}
		foreach insert $ST_INSERTS($wc,$funct) {
		    lappend startpath [car $insert]
		}
		set startpath [newpath [one-or-many $startpath "or"]]
	    }
	    
	    # Now look at combos of presels
	    set results [list [list $startpath [feature-path unit]]]
	    foreach assign $ST_ASSIGNS($wc,$funct) {
		set trigger [car $assign]
		set presel [second $assign]
		
		set newresults {}
		
		foreach itm $results {
		    set ppath [car $itm]
		    set dpath [second $itm]
		    set reals [third $itm]
		    
		    # Add an element for including the presel
		    set i_ppath [simplify-path [require-feature $trigger $ppath]]
		    if { $i_ppath != {} } {
			set reals1 [concat $reals [list $assign]]
			set i_dpath [simplify-path [require-le $presel $dpath]]
			if { $i_dpath == {} } {
			    statictest-insert "\nCombination of preselections of '$funct' failed:"
			    foreach real $reals1 {
				statictest-insert "    [car $real] -> $funct : \{[second $real]\}"
			    }
			} else {
			    lappend newresults [list $i_ppath $i_dpath $reals1]
			}
		    } 
		    
		    # Add an element for excluding the presel
		    set e_ppath [simplify-path [restrict-feature $trigger $ppath]]
		    if { $e_ppath != {} } {
			lappend newresults [list $e_ppath $dpath $reals]
		    } 
		}
		
		set results $newresults
	    }
	    # Complain if no combo worked
	    if { $results == {} } {
		statictest-insert "ERROR: Inserts for $wc / $funct / $slot are not compatible: $ASSIGNS($wc,$funct)"
	    }
	}
    }
}


